// BDE32 3.x - (C) Copyright 1996 by Borland International

// FiltComp.c
#include "snipit.h"

static const char szTblName[] = "customer";
static const char szTblType[] = szPARADOX;
static const CHAR szField[] = "NAME";
static const CHAR szConst[] = "OCEAN";

//==============================================================
//  Function:
//          FiltComp();
//
//  Description:
//          This example shows how to use filters to limit the
//          result set of a table. Specifically, this example
//          does a comparison of the first N digits of a character
//          field. Filters perform a function similar to that of
//          ranges, but more powerful operations are supported.
//==============================================================
void
FiltComp (void)
{
    hDBIDb          hDb = 0;            // Handle to the database.
    hDBICur         hCur = 0;           // Handle to the table.
    DBIResult       rslt;               // Value returned from IDAPI functions
    pBYTE           pcanExpr;           // Structure containing
                                        //   filter info.
    hDBIFilter      hFilter;            // Filter handle.
    UINT16          uSizeNodes;         // Size of the nodes in the
                                        //   tree.
    UINT16          uSizeCanExpr;       // Size of the header
                                        //   information.
    UINT32          uSizeLiterals;      // Size of the literals.
    UINT32          uTotalSize;         // Total size of the filter
                                        //   expression.
    UINT32          uNumRecs  = 10;     // Number of records to
                                        //   display.
    CANExpr         canExp;             // Contains the header
                                        //   information.
    struct {
       CANCompare CompareNode;
       CANField  FieldNode;
       CANConst  ConstantNode;
    } Nodes = {                         // Nodes of the filter tree.
    {
        // Offset 0
        nodeCOMPARE,    // canCompare.nodeClass
        canEQ,          // canCompare.canOp
        1,              // canCompare.bCaseInsensitive
        5,              // canCompare.iParialLen (0 if
                        //   full length)
        sizeof(Nodes.CompareNode),             // canBinary.iOperand1
        sizeof(Nodes.CompareNode) + sizeof(Nodes.FieldNode),
                                // canBinary.iOperand2
    },               //   Offsets in the Nodes array
    {
        // Offset sizeof(Nodes.CompareNode)
        nodeFIELD,      // canField.nodeClass
        canFIELD,       // canField.canOp
        2,              // canField.iFieldNum
        0,              // canField.iNameOffset: szField
    },               //   is the literal at offset 0
    {
        // Offset sizeof(Nodes.CompareNode) + sizeof(Nodes.FieldNode)
        nodeCONST,      // canConst.nodeClass
        canCONST,       // canConst.canOp
        fldZSTRING,     // canConst.iType
        6,              // canConst.iSize strlen("OCEAN") + 1
        5               // canConst.iOffset: szConst is
                               //   the literal at offset
    }};                //   strlen(szField) + 1

    Screen("*** Compare Filter Example ***\r\n");

    BREAK_IN_DEBUGGER();

    Screen("    Initializing IDAPI...");
    if (InitAndConnect(&hDb) != DBIERR_NONE)
    {
        Screen("\r\n*** End of Example ***");
        return;
    }

    Screen("    Setting the database directory...");
    rslt = DbiSetDirectory(hDb, (pCHAR)szTblDirectory);
    ChkRslt(rslt, "SetDirectory");

    Screen("    Open the %s table...", szTblName);
    rslt = DbiOpenTable(hDb, (pCHAR)szTblName, (pCHAR)szTblType, NULL, NULL, 0,
                        dbiREADWRITE, dbiOPENSHARED, xltFIELD, FALSE, NULL,
                        &hCur);
    if (ChkRslt(rslt, "OpenTable") != DBIERR_NONE)
    {
        CloseDbAndExit(&hDb);
        Screen("\r\n*** End of Example ***");
        return;
    }

    // Go to the beginning of the table
    rslt = DbiSetToBegin(hCur);
    ChkRslt(rslt, "SetToBegin");

    Screen("\r\n    Display the %s table...", szTblName);
    DisplayTable(hCur, uNumRecs);

    uSizeNodes      = sizeof(Nodes);   // Size of the nodes.
    // Size of the literals.
    uSizeLiterals   = strlen(szField) + 1 + strlen(szConst) + 1;

    // Size of the header information.
    uSizeCanExpr    = sizeof(CANExpr);
    // Total size of the filter.
    uTotalSize      = uSizeCanExpr + uSizeNodes + uSizeLiterals;

    // Initialize the header information
    canExp.iVer = 1;                         // Version.
    canExp.iTotalSize = (UINT16)uTotalSize;  // Total size of the filter.
    canExp.iNodes = 3;                       // Number of nodes.
    canExp.iNodeStart = uSizeCanExpr;        // The offset in the
                                             //   buffer where the
                                             //   expression nodes
                                             //   start.

    // The offset in the buffer where the literals start.
    //    uSizeCanExprNodes = uSizeCanExpr + uSizeNodes;
    //    canExp.iLiteralStart = (UINT16)uSizeCanExprNodes;
    canExp.iLiteralStart = (UINT16)(uSizeCanExpr + uSizeNodes);

    // Allocate contiguous memory space to hold
    // 1) Header information  i.e. the CANExpr structure
    // 2) Compare, field and constant nodes  i.e. the Nodes structure
    // 3) Literal and constant pool  i.e. field names and constant values
    pcanExpr = (pBYTE)malloc(uTotalSize * sizeof(BYTE));
    if (pcanExpr == NULL)
    {
        Screen("    Could not allocate memory...");
        DbiCloseCursor(&hCur);
        CloseDbAndExit(&hDb);
        Screen("\r\n*** End of Example ***");
        return;
    }

    // Initialize the filter expression by placing header, nodes and
    // pool into pcanexpr

    // Move header information into pcanexpr. pcanExpr will now look as follows:
    // **canExp**|                 |                        |
    // | CANExpr |    All Nodes    | Literal, Constant Pool |
    // |----------------------------------------------------|
    // 0                                                    sizeof(uTotalSize)
    memmove(pcanExpr, &canExp, uSizeCanExpr);

    // Move node structure into pcanexpr.  pcanExpr will now look as follows:
    // |**canExp*|**Node Structure*|                        |
    // | CANExpr |    All Nodes    | Literal, Constant Pool |
    // |----------------------------------------------------|
    // 0                                                    sizeof(uTotalSize)
    memmove(&pcanExpr[uSizeCanExpr], &Nodes, uSizeNodes);

    // Move the literal into pcanexpr.  pcanExpr will now look as follows:
    // |**canExp*|**Node Structure*|***szField              |
    // | CANExpr |    All Nodes    | Literal, Constant Pool |
    // |----------------------------------------------------|
    // 0                                                    sizeof(uTotalSize)
    memmove(&pcanExpr[uSizeCanExpr + uSizeNodes],
            szField, strlen(szField) + 1); // First literal "NAME"

    // Move the literal into pcanexpr.  pcanExpr will now look as follows:
    // |**canExp*|**Node Structure*|***szField*****fConst***|
    // | CANExpr |    All Nodes    | Literal, Constant Pool |
    // |----------------------------------------------------|
    // 0                                                    sizeof(uTotalSize)
    memmove(&pcanExpr[uSizeCanExpr + uSizeNodes + strlen(szField) + 1],
            szConst, strlen(szConst) + 1); // Second literal "OCEAN"

    rslt = DbiSetToBegin(hCur);
    ChkRslt(rslt, "SetToBegin");

    Screen("\r\n    Add a filter to the %s table which will"
           " limit the records\r\n        which are displayed"
           " to those whose %s field starts with '%s'...",
           szTblName, szField, szConst);
    rslt = DbiAddFilter(hCur, 0L, 1, FALSE, (pCANExpr)pcanExpr, NULL, &hFilter);
    if (ChkRslt(rslt, "AddFilter") != DBIERR_NONE)
    {
        rslt = DbiCloseCursor(&hCur);
        ChkRslt(rslt, "CloseCursor");
        free(pcanExpr);
        CloseDbAndExit(&hDb);
        Screen("\r\n*** End of Example ***");
        return;
    }

    // Activate the filter.
    Screen("    Activate the filter on the %s table...",
           szTblName);
    rslt = DbiActivateFilter(hCur, hFilter);
    ChkRslt(rslt, "ActivateFilter");

    rslt = DbiSetToBegin(hCur);
    ChkRslt(rslt, "SetToBegin");

    Screen("\r\n    Display the %s table with the filter"
           " set...", szTblName);
    DisplayTable(hCur, uNumRecs);

    Screen("\r\n    Deactivate the filter...");
    rslt = DbiDeactivateFilter(hCur, hFilter);
    ChkRslt(rslt, "DeactivateFilter");

    Screen("\r\n    Drop the filter...");
    rslt = DbiDropFilter(hCur, hFilter);
    ChkRslt(rslt, "DropFilter");

    rslt = DbiCloseCursor(&hCur);
    ChkRslt(rslt, "CloseCursor");

    free(pcanExpr);

    Screen("    Close the database and exit IDAPI...");
    CloseDbAndExit(&hDb);

    Screen("\r\n*** End of Example ***");
}
